/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2007 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <signal.h>

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <polkit/polkit.h>

static gboolean no_exit = FALSE;

/*--------------------------------------------------------------------------------------------------------------*/

#define GNOME_SYSTEM_MONITOR_TYPE_MECHANISM         (gnome_system_monitor_mechanism_get_type ())
#define GNOME_SYSTEM_MONITOR_MECHANISM(o)           (G_TYPE_CHECK_INSTANCE_CAST ((o), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, GnomeSystemMonitorMechanism))
#define GNOME_SYSTEM_MONITOR_MECHANISM_CLASS(k)     (G_TYPE_CHECK_CLASS_CAST((k), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, GnomeSystemMonitorMechanismClass))
#define GNOME_SYSTEM_MONITOR_IS_MECHANISM(o)        (G_TYPE_CHECK_INSTANCE_TYPE ((o), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM))
#define GNOME_SYSTEM_MONITOR_IS_MECHANISM_CLASS(k)  (G_TYPE_CHECK_CLASS_TYPE ((k), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM))
#define GNOME_SYSTEM_MONITOR_MECHANISM_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, GnomeSystemMonitorMechanismClass))

typedef struct GnomeSystemMonitorMechanismPrivate GnomeSystemMonitorMechanismPrivate;

typedef struct
{
        GObject        parent;
        GnomeSystemMonitorMechanismPrivate *priv;
} GnomeSystemMonitorMechanism;

typedef struct
{
        GObjectClass   parent_class;
} GnomeSystemMonitorMechanismClass;

typedef enum
{
        GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL,
        GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_NOT_PRIVILEGED,
        GNOME_SYSTEM_MONITOR_MECHANISM_NUM_ERRORS
} GnomeSystemMonitorMechanismError;

#define GNOME_SYSTEM_MONITOR_MECHANISM_ERROR gnome_system_monitor_mechanism_error_quark ()

GType gnome_system_monitor_mechanism_error_get_type (void);
#define GNOME_SYSTEM_MONITOR_MECHANISM_TYPE_ERROR (gnome_system_monitor_mechanism_error_get_type ())


GQuark                     gnome_system_monitor_mechanism_error_quark         (void);
GType                      gnome_system_monitor_mechanism_get_type            (void);
GnomeSystemMonitorMechanism *gnome_system_monitor_mechanism_new                 (void);

/* exported methods */
gboolean gnome_system_monitor_mechanism_kill (GnomeSystemMonitorMechanism  *mechanism,
                                              int                           pid,
                                              int                           sig,
                                              DBusGMethodInvocation        *context);

gboolean gnome_system_monitor_mechanism_renice (GnomeSystemMonitorMechanism  *mechanism,
                                                int                           pid,
                                                int                           nice_value,
                                                DBusGMethodInvocation        *context);

/*--------------------------------------------------------------------------------------------------------------*/
#include "gnome-system-monitor-mechanism-glue.h"

static gboolean
do_exit (gpointer user_data)
{
        g_debug ("Exiting due to inactivity");
        exit (1);
        return FALSE;
}

static void
reset_killtimer (void)
{
        static guint timer_id = 0;

        if (no_exit)
                return;

        if (timer_id > 0) {
                g_source_remove (timer_id);
        }
        g_debug ("Setting killtimer to 30 seconds...");
        timer_id = g_timeout_add (30 * 1000, do_exit, NULL);
}

struct GnomeSystemMonitorMechanismPrivate
{
        DBusGConnection *system_bus_connection;
        DBusGProxy      *system_bus_proxy;
        PolkitAuthority *pol_ctx;
};

static void     gnome_system_monitor_mechanism_class_init  (GnomeSystemMonitorMechanismClass *klass);
static void     gnome_system_monitor_mechanism_init        (GnomeSystemMonitorMechanism      *seat);
static void     gnome_system_monitor_mechanism_finalize    (GObject     *object);

G_DEFINE_TYPE (GnomeSystemMonitorMechanism, gnome_system_monitor_mechanism, G_TYPE_OBJECT)

#define GNOME_SYSTEM_MONITOR_MECHANISM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, GnomeSystemMonitorMechanismPrivate))

GQuark
gnome_system_monitor_mechanism_error_quark (void)
{
        static GQuark ret = 0;

        if (ret == 0) {
                ret = g_quark_from_static_string ("gnome_system_monitor_mechanism_error");
        }

        return ret;
}


#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType
gnome_system_monitor_mechanism_error_get_type (void)
{
        static GType etype = 0;
        
        if (etype == 0)
        {
                static const GEnumValue values[] =
                        {
                                ENUM_ENTRY (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL, "GeneralError"),
                                ENUM_ENTRY (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_NOT_PRIVILEGED, "NotPrivileged"),
                                { 0, 0, 0 }
                        };
                
                g_assert (GNOME_SYSTEM_MONITOR_MECHANISM_NUM_ERRORS == G_N_ELEMENTS (values) - 1);
                
                etype = g_enum_register_static ("GnomeSystemMonitorMechanismError", values);
        }
        
        return etype;
}


static GObject *
gnome_system_monitor_mechanism_constructor (GType                  type,
                                            guint                  n_construct_properties,
                                            GObjectConstructParam *construct_properties)
{
        GnomeSystemMonitorMechanism      *mechanism;
        GnomeSystemMonitorMechanismClass *klass;

        klass = GNOME_SYSTEM_MONITOR_MECHANISM_CLASS (g_type_class_peek (GNOME_SYSTEM_MONITOR_TYPE_MECHANISM));

        mechanism = GNOME_SYSTEM_MONITOR_MECHANISM (
                G_OBJECT_CLASS (gnome_system_monitor_mechanism_parent_class)->constructor (type,
                                                                                           n_construct_properties,
                                                                                           construct_properties));
        
        return G_OBJECT (mechanism);
}

static void
gnome_system_monitor_mechanism_class_init (GnomeSystemMonitorMechanismClass *klass)
{
        GObjectClass   *object_class = G_OBJECT_CLASS (klass);

        object_class->constructor = gnome_system_monitor_mechanism_constructor;
        object_class->finalize = gnome_system_monitor_mechanism_finalize;

        g_type_class_add_private (klass, sizeof (GnomeSystemMonitorMechanismPrivate));

        dbus_g_object_type_install_info (GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, &dbus_glib_gnome_system_monitor_mechanism_object_info);

        dbus_g_error_domain_register (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR, NULL, GNOME_SYSTEM_MONITOR_MECHANISM_TYPE_ERROR);

}

static void
gnome_system_monitor_mechanism_init (GnomeSystemMonitorMechanism *mechanism)
{
        mechanism->priv = GNOME_SYSTEM_MONITOR_MECHANISM_GET_PRIVATE (mechanism);

}

static void
gnome_system_monitor_mechanism_finalize (GObject *object)
{
        GnomeSystemMonitorMechanism *mechanism;

        g_return_if_fail (object != NULL);
        g_return_if_fail (GNOME_SYSTEM_MONITOR_IS_MECHANISM (object));

        mechanism = GNOME_SYSTEM_MONITOR_MECHANISM (object);

        g_return_if_fail (mechanism->priv != NULL);

        g_object_unref (mechanism->priv->system_bus_proxy);

        G_OBJECT_CLASS (gnome_system_monitor_mechanism_parent_class)->finalize (object);
}

static gboolean
register_mechanism (GnomeSystemMonitorMechanism *mechanism)
{
        GError *error = NULL;

        mechanism->priv->pol_ctx = polkit_authority_get ();

        error = NULL;
        mechanism->priv->system_bus_connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (mechanism->priv->system_bus_connection == NULL) {
                if (error != NULL) {
                        g_critical ("error getting system bus: %s", error->message);
                        g_error_free (error);
                }
                goto error;
        }

        dbus_g_connection_register_g_object (mechanism->priv->system_bus_connection, "/", 
                                             G_OBJECT (mechanism));

        mechanism->priv->system_bus_proxy = dbus_g_proxy_new_for_name (mechanism->priv->system_bus_connection,
                                                                      DBUS_SERVICE_DBUS,
                                                                      DBUS_PATH_DBUS,
                                                                      DBUS_INTERFACE_DBUS);

        reset_killtimer ();

        return TRUE;

error:
        return FALSE;
}


GnomeSystemMonitorMechanism *
gnome_system_monitor_mechanism_new (void)
{
        GObject *object;
        gboolean res;

        object = g_object_new (GNOME_SYSTEM_MONITOR_TYPE_MECHANISM, NULL);

        res = register_mechanism (GNOME_SYSTEM_MONITOR_MECHANISM (object));
        if (! res) {
                g_object_unref (object);
                return NULL;
        }

        return GNOME_SYSTEM_MONITOR_MECHANISM (object);
}

static PolkitSubject *
get_caller_from_context (GnomeSystemMonitorMechanism *mechanism, DBusGMethodInvocation *context)
{
        char *sender;
        PolkitSubject *pk_caller;

        sender = dbus_g_method_get_sender (context);
        pk_caller = polkit_system_bus_name_new (sender);
	g_free (sender);

        return pk_caller;
}

/* note, we take ownership of the passed PolkitSubject object */
static gboolean
_check_polkit_for_action_for_caller (GnomeSystemMonitorMechanism *mechanism,
                                     DBusGMethodInvocation *context,
                                     PolkitSubject *subject,
                                     const char *action)
{
        GError *error;
        PolkitAuthorizationResult *result;

        error = NULL;

        result = polkit_authority_check_authorization_sync (mechanism->priv->pol_ctx,
                                                            subject,
                                                            action,
                                                            NULL,
                                                            POLKIT_CHECK_AUTHORIZATION_FLAGS_ALLOW_USER_INTERACTION,
                                                            NULL, NULL);
        g_object_unref (subject);

        if (!polkit_authorization_result_get_is_authorized (result)) {
                error = g_error_new (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR,
                                     GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_NOT_PRIVILEGED,
                                     "%s",
                                     action);
                dbus_g_method_return_error (context, error);
                g_error_free (error);
		g_object_unref (result);
                return FALSE;
        }

	g_object_unref (result);
        return TRUE;
}

static gboolean
_check_polkit_for_action (GnomeSystemMonitorMechanism *mechanism,
                          DBusGMethodInvocation *context,
                          const char *action)
{
        PolkitSubject *subject;

        /* Check that caller is privileged */
        if ((subject = get_caller_from_context (mechanism, context)) == NULL)
                return FALSE;

        return _check_polkit_for_action_for_caller (mechanism, context, subject, action);
}

/*--------------------------------------------------------------------------------------------------------------*/
/* exported methods */

gboolean
gnome_system_monitor_mechanism_kill (GnomeSystemMonitorMechanism  *mechanism,
                                     int                           pid,
                                     int                           sig,
                                     DBusGMethodInvocation        *context)
{
        reset_killtimer ();

        if (!_check_polkit_for_action (mechanism, context, "org.gnome.system-monitor.kill"))
                return FALSE;

        g_debug ("Sending signal %d to pid %d", sig, pid);

        if (kill (pid, sig) != 0) {
                GError *error;
                error = g_error_new (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR,
                                     GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL,
                                     "Error sending signal %d to pid %d: %m", sig, pid);
                g_warning ("Error: %s", error->message);
                dbus_g_method_return_error (context, error);
                g_error_free (error);
                return FALSE;
        }

        dbus_g_method_return (context);
        return TRUE;
}

static gboolean
get_uid_from_context (GnomeSystemMonitorMechanism  *mechanism,
                      DBusGMethodInvocation        *context,
                      uid_t                        *uid)
{
        char *sender;
        DBusConnection *con;

        con = dbus_g_connection_get_connection (dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL));
        if (con == NULL)
                return FALSE;

        sender = dbus_g_method_get_sender (context);
	*uid = dbus_bus_get_unix_user (con, sender, NULL);
        g_free (sender);

        if (*uid == (unsigned)-1)
                return FALSE;

        return TRUE;
}

gboolean
gnome_system_monitor_mechanism_renice (GnomeSystemMonitorMechanism  *mechanism,
                                       int                           pid,
                                       int                           nice_value,
                                       DBusGMethodInvocation        *context)
{
        uid_t caller_uid;
        uid_t process_uid;
        PolkitSubject *subject;
        char *procpath;
        struct stat statbuf;
        const char *action_id;

        reset_killtimer ();

        /* Hmm; is there a better way to do this? This may be Linux specific.. */
        procpath = g_strdup_printf ("/proc/%d", pid);
        if (stat (procpath, &statbuf) != 0) {
                GError *error;
                error = g_error_new (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR,
                                     GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL,
                                     "Error figuring out uid for pid %d", pid);
                g_warning ("Error: %s", error->message);
                dbus_g_method_return_error (context, error);
                g_error_free (error);
                g_free (procpath);
                return FALSE;
        }
        g_free (procpath);
        process_uid = statbuf.st_uid;

        subject = get_caller_from_context (mechanism, context);
        if (subject == NULL)
                return FALSE;

        if (!get_uid_from_context (mechanism, context, &caller_uid)) {
                GError *error;
                error = g_error_new (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR,
                                     GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL,
                                     "Error figuring out uid of caller");
                g_warning ("Error: %s", error->message);
                dbus_g_method_return_error (context, error);
                g_error_free (error);
                g_object_unref (subject);
                return FALSE;
        }

        if (caller_uid == process_uid) {
                action_id = "org.gnome.system-monitor.increase-own-priority";
        } else {
                action_id = "org.gnome.system-monitor.change-priority";
        }

        if (!_check_polkit_for_action_for_caller (mechanism, context, subject, action_id)) {
                return FALSE;
        }

        g_debug ("Renicing pid %d (owned by uid %d) to priority %d on behalf of uid %d (action_id=%s)",
                 pid, process_uid, nice_value, caller_uid, action_id);

        errno = 0;
        if (setpriority (PRIO_PROCESS, pid, nice_value) != 0) {
                GError *error;
                error = g_error_new (GNOME_SYSTEM_MONITOR_MECHANISM_ERROR,
                                     GNOME_SYSTEM_MONITOR_MECHANISM_ERROR_GENERAL,
                                     "Error renicing pid %d to priority %d: %m", pid, nice_value);
                g_warning ("Error: %s", error->message);
                dbus_g_method_return_error (context, error);
                g_error_free (error);
                return FALSE;
        }

        dbus_g_method_return (context);
        return TRUE;
}

/*--------------------------------------------------------------------------------------------------------------*/

#define BUS_NAME "org.gnome.SystemMonitor.Mechanism"

static gboolean
acquire_name_on_proxy (DBusGProxy *bus_proxy)
{
        GError     *error;
        guint       result;
        gboolean    res;
        gboolean    ret;

        ret = FALSE;

        if (bus_proxy == NULL) {
                goto out;
        }

        error = NULL;
	res = dbus_g_proxy_call (bus_proxy,
                                 "RequestName",
                                 &error,
                                 G_TYPE_STRING, BUS_NAME,
                                 G_TYPE_UINT, 0,
                                 G_TYPE_INVALID,
                                 G_TYPE_UINT, &result,
                                 G_TYPE_INVALID);
        if (! res) {
                if (error != NULL) {
                        g_warning ("Failed to acquire %s: %s", BUS_NAME, error->message);
                        g_error_free (error);
                } else {
                        g_warning ("Failed to acquire %s", BUS_NAME);
                }
                goto out;
	}

 	if (result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
                if (error != NULL) {
                        g_warning ("Failed to acquire %s: %s", BUS_NAME, error->message);
                        g_error_free (error);
                } else {
                        g_warning ("Failed to acquire %s", BUS_NAME);
                }
                goto out;
        }

        ret = TRUE;

 out:
        return ret;
}

int
main (int argc, char **argv)
{
        GMainLoop                    *loop;
        GnomeSystemMonitorMechanism  *mechanism;
        DBusGProxy                   *bus_proxy;
        DBusGConnection              *connection;
        int                           ret;
        GError                       *error;
        GOptionContext      *context;
        static GOptionEntry entries []   = {
                { "no-exit", 0, 0, G_OPTION_ARG_NONE, &no_exit, N_("Don't exit after 30 seconds of inactivity"), NULL },
                { NULL }
        };

        ret = 1;

        g_type_init ();

        context = g_option_context_new (_("PolicyKit GNOME session daemon"));
        g_option_context_add_main_entries (context, entries, NULL);
        g_option_context_parse (context, &argc, &argv, NULL);
        g_option_context_free (context);


        error = NULL;
        connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (connection == NULL) {
                g_warning ("Couldn't connect to system bus: %s", error->message);
                g_error_free (error);
                goto out;
        }

	bus_proxy = dbus_g_proxy_new_for_name (connection,
                                               DBUS_SERVICE_DBUS,
                                               DBUS_PATH_DBUS,
                                               DBUS_INTERFACE_DBUS);
        if (bus_proxy == NULL) {
                g_warning ("Could not construct bus_proxy object; bailing out");
                goto out;
        }

        if (!acquire_name_on_proxy (bus_proxy) ) {
                g_warning ("Could not acquire name; bailing out");
                goto out;
        }

        mechanism = gnome_system_monitor_mechanism_new ();

        if (mechanism == NULL) {
                goto out;
        }

        loop = g_main_loop_new (NULL, FALSE);

        g_main_loop_run (loop);

        g_object_unref (mechanism);
        g_main_loop_unref (loop);
        ret = 0;

out:
        return ret;
}
