/* gsm-session-manager.h - functions for monitoring the session
 *
 * Copyright (C) 2008 Red Hat, Inc
 * Author: Matthias Clasen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <dbus/dbus-glib.h>
#include <gsm-session-manager.h>
#include <string.h>

struct _GsmSessionManagerPrivate
{
	DBusGConnection  *dbus_connection;
        DBusGProxy       *bus_proxy;
        DBusGProxy       *sm_proxy;
	guint		  is_connected :1;

	GHashTable	 *clients;
};

static void gsm_session_manager_finalize                 (GObject *object);
static void gsm_session_manager_class_install_signals    (GsmSessionManagerClass *manager_class);

static void gsm_session_manager_class_install_properties (GsmSessionManagerClass *manager_class);

static void gsm_session_manager_get_property (GObject    *object,
                                              guint       prop_id,
                                              GValue     *value,
                                              GParamSpec *pspec);

static gboolean ensure_connection            (GsmSessionManager  *manager,
                                              GError            **error);
static void     get_all_clients              (GsmSessionManager *manager);

enum {
        PROP_0 = 0,
        PROP_IS_CONNECTED
};

enum {
	CLIENT_ADDED = 0,
	CLIENT_REMOVED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

G_DEFINE_TYPE (GsmSessionManager, gsm_session_manager, G_TYPE_OBJECT);

static void
gsm_session_manager_class_init (GsmSessionManagerClass *manager_class)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (manager_class);

	object_class->finalize = gsm_session_manager_finalize;

        gsm_session_manager_class_install_properties (manager_class);
        gsm_session_manager_class_install_signals (manager_class);

	g_type_class_add_private (manager_class,
				  sizeof (GsmSessionManagerPrivate));
}

static void
gsm_session_manager_class_install_signals (GsmSessionManagerClass *manager_class)
{
        GObjectClass *object_class;

        object_class = G_OBJECT_CLASS (manager_class);

	signals[CLIENT_ADDED] =
		g_signal_new ("client-added",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (GsmSessionManagerClass, client_added),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_STRING);
	signals[CLIENT_REMOVED] =
		g_signal_new ("client-removed",
			      G_OBJECT_CLASS_TYPE (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (GsmSessionManagerClass, client_removed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_STRING);
}

static void
gsm_session_manager_class_install_properties (GsmSessionManagerClass *manager_class)
{
        GObjectClass *object_class;
        GParamSpec   *param_spec;

        object_class = G_OBJECT_CLASS (manager_class);
        object_class->get_property = gsm_session_manager_get_property;

        param_spec = g_param_spec_boolean ("is-connected",
                                           "Is connected",
                                           "Whether we are connected to "
                                           "the session manager",
                                           FALSE,
                                           G_PARAM_READABLE);

        g_object_class_install_property (object_class, PROP_IS_CONNECTED,
                                         param_spec);
}

static void
on_name_owner_changed (DBusGProxy        *bus_proxy,
                       const char        *name,
                       const char        *prev_owner,
                       const char        *new_owner,
                       GsmSessionManager *manager)
{
        if (strcmp (name, "org.gnome.SessionManager") != 0) {
                return;
        }

        if (manager->priv->sm_proxy != NULL) {
                g_object_unref (manager->priv->sm_proxy);
                manager->priv->sm_proxy = NULL;
        }

	g_hash_table_remove_all (manager->priv->clients);

        if (ensure_connection (manager, NULL))
		get_all_clients (manager);
}


static void
get_client_details (GsmSessionManager        *manager,
		    const char               *obj_path,
                    GsmClientDetailsCallback  callback,
                    gpointer                  data)
{
        DBusGProxy *proxy;
        GError *error = NULL;
        gboolean ret;
        guint status;
        guint restart_style_hint;
        guint pid;

        proxy = dbus_g_proxy_new_for_name (manager->priv->dbus_connection,
                                           "org.gnome.SessionManager",
                                           obj_path,
                                           "org.gnome.SessionManager.Client");

        ret = dbus_g_proxy_call (proxy, "GetUnixProcessId", &error,
                                 G_TYPE_INVALID,
                                 G_TYPE_UINT, &pid,
                                 G_TYPE_INVALID);
        if (error) {
                g_warning ("ERROR: %s\n", error->message);
                g_error_free (error);
                goto cleanup;
        }
        if (!ret) {
                g_warning ("GetUnixProcessId failed\n");
                goto cleanup;
        }

        ret = dbus_g_proxy_call (proxy, "GetStatus", &error,
                                 G_TYPE_INVALID,
                                 G_TYPE_UINT, &status,
                                 G_TYPE_INVALID);
        if (error) {
                g_warning ("ERROR: %s\n", error->message);
                g_error_free (error);
                goto cleanup;
        }
        if (!ret) {
                g_warning ("GetStatus failed\n");
                goto cleanup;
        }

        ret = dbus_g_proxy_call (proxy, "GetRestartStyleHint", &error,
                                 G_TYPE_INVALID,
                                 G_TYPE_UINT, &restart_style_hint,
                                 G_TYPE_INVALID);
        if (error) {
                g_warning ("ERROR: %s\n", error->message);
                g_error_free (error);
                goto cleanup;
        }
        if (!ret) {
                g_warning ("GetRestartStyleHint failed\n");
                goto cleanup;
        }

        callback (obj_path, pid, status, restart_style_hint, data);

cleanup:
        g_object_unref (proxy);
}


typedef struct {
	guint       pid;
	guint	    status;
	guint       restart_style_hint;
} ClientDetails;

void
gsm_session_manager_foreach (GsmSessionManager        *manager,
                             GsmClientDetailsCallback  callback,
                             gpointer                  data)
{
	GHashTableIter iter;
	gchar *key;
	ClientDetails *value;

	g_hash_table_iter_init (&iter, manager->priv->clients);
	while (g_hash_table_iter_next (&iter,(gpointer *)&key, (gpointer *)&value)) {
		callback (key, value->pid, value->status, value->restart_style_hint, data);
	}
}

static void
add_client (const char        *obj_path,
	    guint              pid,
            guint              status,
            guint              restart_style_hint,
	    gpointer           data)
{
	GsmSessionManager *manager = data;

	ClientDetails *details;

	details = g_new (ClientDetails, 1);
	details->pid = pid;
	details->status = status;
	details->restart_style_hint = restart_style_hint;

	g_hash_table_insert (manager->priv->clients, 
			     g_strdup (obj_path),
			     details);	
}

static void
get_all_clients (GsmSessionManager *manager)
{
        GError *error = NULL;
        GPtrArray *array;
        gboolean ret;
        gchar *obj_path;
        guint i;

        ret = dbus_g_proxy_call (manager->priv->sm_proxy, 
				 "GetClients", &error,
                                 G_TYPE_INVALID,
                                 dbus_g_type_get_collection ("GPtrArray", DBUS_TYPE_G_OBJECT_PATH), &array,
                                 G_TYPE_INVALID);

        if (error) {
                g_warning ("ERROR: %s\n", error->message);
                g_error_free (error);
                return;
        }
        if (!ret) {
                g_warning ("GetClients failed\n");
        }

        for (i = 0; i < array->len; i++) {
                obj_path = g_ptr_array_index (array, i);
                get_client_details (manager, obj_path, add_client, manager);
        }

        g_ptr_array_foreach (array, (GFunc)g_free, NULL);
        g_ptr_array_free (array, TRUE);
}



static void
on_client_added (DBusGProxy        *sm_proxy,
		 const gchar       *obj_path,
		 GsmSessionManager *manager)
{
        get_client_details (manager, obj_path, add_client, manager);

	g_signal_emit (manager, signals[CLIENT_ADDED], 0, obj_path);
}

static void
on_client_removed (DBusGProxy        *sm_proxy,
                   const char        *obj_path,
                   GsmSessionManager *manager)
{
	g_signal_emit (manager, signals[CLIENT_REMOVED], 0, obj_path);
	
	g_hash_table_remove (manager->priv->clients, obj_path);
}

static gboolean
ensure_connection (GsmSessionManager  *manager,
                   GError            **error)
{
        GError   *connection_error;
        gboolean  is_connected;

        connection_error = NULL;
        if (manager->priv->dbus_connection == NULL) {
                manager->priv->dbus_connection = dbus_g_bus_get (DBUS_BUS_SESSION,
                                                                 &connection_error);

                if (manager->priv->dbus_connection == NULL) {
                        g_propagate_error (error, connection_error);
                        is_connected = FALSE;
                        goto out;
                }
        }
        if (manager->priv->bus_proxy == NULL) {
                manager->priv->bus_proxy =
                        dbus_g_proxy_new_for_name_owner (manager->priv->dbus_connection,
                                                         DBUS_SERVICE_DBUS,
                                                         DBUS_PATH_DBUS,
                                                         DBUS_INTERFACE_DBUS,
                                                         &connection_error);

                if (manager->priv->bus_proxy == NULL) {
                        g_propagate_error (error, connection_error);
                        is_connected = FALSE;
                        goto out;
                }

                dbus_g_proxy_add_signal (manager->priv->bus_proxy,
                                         "NameOwnerChanged",
                                         G_TYPE_STRING,
                                         G_TYPE_STRING,
                                         G_TYPE_STRING,
                                         G_TYPE_INVALID);
                dbus_g_proxy_connect_signal (manager->priv->bus_proxy,
                                             "NameOwnerChanged",
                                             G_CALLBACK (on_name_owner_changed),
                                             manager, NULL);
        }

        if (manager->priv->sm_proxy == NULL) {
                manager->priv->sm_proxy =
                        dbus_g_proxy_new_for_name_owner (manager->priv->dbus_connection,
                                                         "org.gnome.SessionManager",
                                                         "/org/gnome/SessionManager",
                                                         "org.gnome.SessionManager",
                                                         &connection_error);

                if (manager->priv->sm_proxy == NULL) {
                        g_propagate_error (error, connection_error);
                        is_connected = FALSE;
                        goto out;
                }

        	dbus_g_proxy_add_signal (manager->priv->sm_proxy, 
					 "ClientAdded", 
					 DBUS_TYPE_G_OBJECT_PATH, 
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (manager->priv->sm_proxy, 
					     "ClientAdded", 
					     G_CALLBACK (on_client_added), 
					     manager, NULL);

        	dbus_g_proxy_add_signal (manager->priv->sm_proxy, 
					 "ClientRemoved", 
					 DBUS_TYPE_G_OBJECT_PATH, 
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (manager->priv->sm_proxy, 
					     "ClientRemoved", 
					     G_CALLBACK (on_client_removed), 
					     manager, NULL);
        }

        is_connected = TRUE;

 out:
        if (manager->priv->is_connected != is_connected) {
                manager->priv->is_connected = is_connected;
                g_object_notify (G_OBJECT (manager), "is-connected");
        }

        if (!is_connected) {
                if (manager->priv->dbus_connection == NULL) {
                        if (manager->priv->bus_proxy != NULL) {
                                g_object_unref (manager->priv->bus_proxy);
                                manager->priv->bus_proxy = NULL;
                        }

                        if (manager->priv->sm_proxy != NULL) {
                                g_object_unref (manager->priv->sm_proxy);
                                manager->priv->sm_proxy = NULL;
                        }
                } else if (manager->priv->bus_proxy == NULL) {
                        if (manager->priv->sm_proxy != NULL) {
                                g_object_unref (manager->priv->sm_proxy);
                                manager->priv->sm_proxy = NULL;
                        }
                }
        }

        return is_connected;
}


static void
gsm_session_manager_init (GsmSessionManager *manager)
{
	GError *error;

	manager->priv = G_TYPE_INSTANCE_GET_PRIVATE (manager, GSM_TYPE_SESSION_MANAGER,
		 				     GsmSessionManagerPrivate);

	manager->priv->dbus_connection = NULL;
	manager->priv->bus_proxy = NULL;
	manager->priv->sm_proxy = NULL;
	manager->priv->is_connected = FALSE;

	manager->priv->clients = g_hash_table_new_full (g_str_hash, g_str_equal,
							g_free, g_free);
	error = NULL;

        if (ensure_connection (manager, &error)) {
		get_all_clients (manager);
	}
	else {
                g_message ("Could not connect to session manager: %s",
                           error->message);
                g_error_free (error);
        }
}

static void
gsm_session_manager_finalize (GObject *object)
{
        GsmSessionManager *manager;
        GObjectClass      *parent_class;

        manager = GSM_SESSION_MANAGER (object);

	g_hash_table_destroy (manager->priv->clients);

        parent_class = G_OBJECT_CLASS (gsm_session_manager_parent_class);

        if (parent_class->finalize != NULL) {
                parent_class->finalize (object);
        }
}

static void
gsm_session_manager_get_property (GObject    *object,
                                  guint       prop_id,
                                  GValue     *value,
                                  GParamSpec *pspec)
{
        GsmSessionManager *manager = GSM_SESSION_MANAGER (object);

        switch (prop_id) {
        case PROP_IS_CONNECTED:
                g_value_set_boolean (value,
                                     manager->priv->is_connected);
                break;

        default:
                G_OBJECT_WARN_INVALID_PROPERTY_ID (object,
                                                   prop_id,
                                                   pspec);
        }
}

GsmSessionManager *
gsm_session_manager_new (void)
{
        GsmSessionManager *manager;

        manager = g_object_new (GSM_TYPE_SESSION_MANAGER, NULL);

        return manager;
}

GsmSessionManager *
gsm_session_manager_get (void)
{
        static GsmSessionManager *manager = NULL;

        if (manager == NULL) {
                manager = gsm_session_manager_new ();
        }

        return g_object_ref (manager);
}

gboolean
gsm_session_manager_get_details (GsmSessionManager *manager,
				 const char        *obj_path,
				 guint             *pid,
				 guint 		   *status,
				 guint		   *restart_style_hint)
{
	ClientDetails *details;

	details = (ClientDetails *)g_hash_table_lookup (manager->priv->clients, obj_path);

	if (details) {
		*pid = details->pid;
		*status = details->status;
		*restart_style_hint = details->restart_style_hint;
		return TRUE;
	}
	
	return FALSE;
}
 
